"""Console script for pyscihub."""
import sys
import logging
import click
from time import sleep
from random import gauss

from .pyscihub import SciHub


@click.group()
@click.option(
    "--output",
    "-o",
    help="Output path for PDFs",
    default="./output/",
    type=click.Path(exists=True),
)
@click.option("--verbose", is_flag=True)
@click.pass_context
def cli(ctx, output, verbose):
    """CLI to download PDFs from Sci-Hub."""
    ctx.ensure_object(dict)
    ctx.obj["OUTPUT"] = output
    ctx.obj["VERBOSE"] = verbose


@cli.command("file")
@click.argument("file_path", type=click.Path(exists=True))
@click.pass_context
def make_file(ctx, file_path):
    scihub = SciHub("https://sci-hub.se", ctx.obj["OUTPUT"])

    # open file
    with open(file_path, "r") as f:
        queries = f.readlines()

    scihub.download(queries)


@cli.command("single")
@click.argument("query", type=str)
@click.pass_context
def make_query(ctx, query):
    scihub = SciHub("https://sci-hub.se", ctx.obj["OUTPUT"])
    scihub.download(query)


def main():
    cli(obj={})


if __name__ == "__main__":
    main()  # pragma: no cover



#!/usr/bin/env python

"""Tests for `pyscihub` package."""


import pytest
import re
import shutil
import csv
from click.testing import CliRunner
from pathlib import Path

from pyscihub import pyscihub
from pyscihub import cli
from pyscihub import tools


TEST_REFERENCES = [
    {
        "query": "A heuristic algorithm for a single vehicle static bike sharing rebalancing problem",
        "authors": None,
        "title": "A heuristic algorithm for a single vehicle static bike sharing rebalancing problem",
    },
    {
        "query": "Iterated local search: Framework and applications",
        "authors": None,
        "title": "Iterated local search: Framework and applications",
    },
    {
        "query": "Fong, C.Y., et al., Chloral hydrate as a sedating agent for neurodiagnostic procedures in children. Cochrane Database of Systematic Reviews, 2017. 2017(11).",
        "authors": "Fong, C.Y., et al.",
        "title": "Chloral hydrate as a sedating agent for neurodiagnostic procedures in children.",
    },
    {
        "query": "Forbes, H., et al., The Effects of Group Membership on College Students' Social Exclusion of Peers and Bystander Behavior. Journal of Psychology, 2020. 154(1): p. 15-37.",
        "authors": "Forbes, H., et al.",
        "title": "The Effects of Group Membership on College Students' Social Exclusion of Peers and Bystander Behavior.",
    },
    {
        "query": 'Force, S., The "innocent bystander" complications following esophagectomy: Atrial fibrillation, recurrent laryngeal nerve injury, chylothorax, and pulmonary complications. Seminars in Thoracic and Cardiovascular Surgery, 2004. 16(2): p. 117-123.',
        "authors": "Force, S.",
        "title": 'The "innocent bystander" complications following esophagectomy: Atrial fibrillation, recurrent laryngeal nerve injury, chylothorax, and pulmonary complications.',
    },
    {
        "query": "Forfar, J.C., D.C. Russell, and M.F. Oliver, Haemodynamic effects of sulphinpyrazone on exercise responses in normal subjects. Lancet, 1980. 2(8197): p. 718-20.",
        "authors": "Forfar, J.C., D.C. Russell, and M.F. Oliver",
        "title": "Haemodynamic effects of sulphinpyrazone on exercise responses in normal subjects.",
    },
    {
        "query": "Fornes, P. and D. Lecomte, Pathology of sport-related sudden death. [French]. Revue du Praticien, 2001. 51(SPEC.ISS): p. 31-35.",
        "authors": "Fornes, P. and D. Lecomte",
        "title": "Pathology of sport-related sudden death.",
    },
    {
        "query": "Freedenberg, V.A., P.S. Hinds, and E. Friedmann, Mindfulness-Based Stress Reduction and Group Support Decrease Stress in Adolescents with Cardiac Diagnoses: A Randomized Two-Group Study. Pediatric Cardiology, 2017. 38(7): p. 1415-1425.",
        "authors": "Freedenberg, V.A., P.S. Hinds, and E. Friedmann",
        "title": "Mindfulness-Based Stress Reduction and Group Support Decrease Stress in Adolescents with Cardiac Diagnoses: A Randomized Two-Group Study.",
    },
]

TEST_QUERIES = [
    {
        "query": "Lourenço, H. R., Martin, O. C., & Stützle, T. (2010). Iterated Local Search: Framework and Applications. International Series in Operations Research & Management Science, 363–397. doi:10.1007/978-1-4419-1665-5_12",
        "clean_query": "10.1007/978-1-4419-1665-5_12",
    },
    {
        "query": "Cruz, F., Subramanian, A., Bruck, B. P., & Iori, M. (2017). A heuristic algorithm for a single vehicle static bike sharing rebalancing problem. Computers & Operations Research, 79, 19–33. doi:10.1016/j.cor.2016.09.025",
        "clean_query": "10.1016/j.cor.2016.09.025",
    },
    {
        "query": "https://www.sciencedirect.com/science/article/abs/pii/S0305054816302489",
        "clean_query": "https://www.sciencedirect.com/science/article/abs/pii/S0305054816302489",
    },
]


@pytest.mark.parametrize("test_ref", TEST_REFERENCES)
def test_reference_regex(test_ref):
    """Test if regex for references correctly captures the title (and authors)."""

    # set REGEX
    REF_REGEX = tools.ref_regex_simple()

    mo = re.search(REF_REGEX, test_ref["query"])
    assert mo.group("authors") == test_ref["authors"]
    assert mo.group("title") == test_ref["title"]


@pytest.mark.parametrize("test_query", TEST_QUERIES)
def test_doi_and_url_regex(test_query):
    """Test that query parser correctly captures DOIs and URLs."""
    assert test_query["clean_query"] == tools.extract_valid_query(test_query["query"])


@pytest.fixture
def init_empty_scihub(tmpdir):
    test_output_path = tmpdir.mkdir("output")
    return pyscihub.SciHub("https://sci-hub.se", test_output_path.realpath())


@pytest.fixture
def init_populated_scihub(tmpdir):
    # create new folder and move data from tests/data/demo_pdfs
    test_output_path = Path(tmpdir.mkdir("output"))
    demo_path = Path("data/demo_pdfs")
    demo_files = [
        demo_file for demo_file in Path(demo_path).iterdir() if demo_file.is_file()
    ]
    for demo_file in demo_files:
        shutil.copy2(demo_file, test_output_path)

    # initiate scihub
    scihub = pyscihub.SciHub("https://sci-hub.se", str(test_output_path.absolute()))

    # change file location of pdfs
    lines = list(csv.reader(open(test_output_path / "pdf_paths.csv", "r")))
    for line in lines[1:]:
        print(line[1])
        line[1] = f"{tmpdir.realpath()}/{line[1]}"
        print(line[1])
    writer = csv.writer(open(test_output_path / "pdf_paths.csv", "w"))
    writer.writerows(lines)

    return scihub


def test_download_arg_type(init_empty_scihub):
    """Test something."""
    scihub = init_empty_scihub

    with pytest.raises(ValueError) as excinfo:
        scihub.download(set("abc"))
        assert "queries argument should be a list or a single string." in excinfo.value


def test_empty_pdf_path(init_empty_scihub):
    """Test that pdf_paths dictionary is indeed empty."""
    scihub = init_empty_scihub

    pdf_paths = scihub._get_pdf_paths()
    assert type(pdf_paths) == dict
    assert len(pdf_paths) == 0


def test_populated_pdf_paths(init_populated_scihub):
    """Test that pdf_paths contains pdf_paths of entries in CSV AND actual files."""
    scihub = init_populated_scihub

    pdf_paths = scihub._get_pdf_paths()
    print(pdf_paths)
    assert pdf_paths[
        "A heuristic algorithm for a single vehicle static bike sharing rebalancing problem"
    ]
    assert pdf_paths[
        "Forbes, H., et al., The Effects of Group Membership on College Students' Social Exclusion of Peers and Bystander Behavior. Journal of Psychology, 2020. 154(1): p. 15-37."
    ]
    with pytest.raises(KeyError) as excinfo:
        pdf_paths[
            "Foo, B., et al., hERG quality control and the long QT syndrome. Journal of Physiology, 2016. 594(9): p. 2469-81."
        ]
        assert (
            "KeyError: 'Foo, B., et al., hERG quality control and the long QT syndrome. Journal of Physiology, 2016. 594(9): p. 2469-81.'"
            in excinfo.value
        )


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()

    # check if main program shows all commands
    result = runner.invoke(cli.cli)
    assert result.exit_code == 0
    assert "cli [OPTIONS] COMMAND [ARGS]..." in result.output
    assert "file" in result.output
    assert "single" in result.output

    # check if help flag is properly configured
    help_result = runner.invoke(cli.cli, ["--help"])
    assert help_result.exit_code == 0
    assert "Show this message" in help_result.output



"""Tools for handling queries."""

import re
import os
import logging


def ref_regex():
    """Return regex for valid reference."""
    FIRST_AUTHOR = r"(?P<first_author>[A-Z][\w\-']+(?:,\s(?:[A-Z]\.)+))"
    SECOND_AUTHOR = r"(?P<other_author>(?:[A-Z]\.)+\s[A-Z][\w\-']+)"
    ET_AL = r"et al\."
    BETWEEN_AUTHORS = r"(?:(?:[,]?\sand\s)|(?:\,\s))"
    ALL_AUTHORS = (
        r"(?:"
        + FIRST_AUTHOR
        + r"(?:(?:"
        + BETWEEN_AUTHORS
        + SECOND_AUTHOR
        + r")+|(?:"
        + BETWEEN_AUTHORS
        + ET_AL
        + r"))?\,\s)?"
    )  # optional

    TITLE = r"(?P<title>[^.]+)"
    JOURNAL = r"(?:\s(?P<journal>[^.]+)[,.])?"  # optional
    YEAR = r"(?:\s(?P<year>(19|20)\d{2})(?:[^.]+)[,.])?"  # optional
    VOLUME = r"(?:\s(?P<volume>\d+(\(\d+\))))?"  # optional
    PAGE = r"(?:\: p\. (?P<page>[0-9e.]+(?:\-[0-9e.]+)?))?"  # optional

    return ALL_AUTHORS + TITLE + JOURNAL + YEAR + VOLUME + PAGE


def ref_regex_simple():
    """Return regex query to extract authors and title from reference."""
    FIRST_AUTHOR = r"(?P<first_author>[A-Z][\w\-']+(?:,\s(?:[A-Z]\.)+))"
    SECOND_AUTHOR = r"(?P<other_author>(?:[A-Z]\.)+\s[A-Z][\w\-']+)"
    ET_AL = r"et al\."
    BETWEEN_AUTHORS = r"(?:(?:[,]?\sand\s)|(?:\,\s))"
    ALL_AUTHORS = (
        r"(?:(?P<authors>"
        + FIRST_AUTHOR
        + r"(?:(?:"
        + BETWEEN_AUTHORS
        + SECOND_AUTHOR
        + r")+|(?:"
        + BETWEEN_AUTHORS
        + ET_AL
        + r"))?)\,\s)?"
    )  # optional
    TITLE = r"(?P<title>[^.]+[\.]?)"

    return r"^" + ALL_AUTHORS + TITLE


def extract_valid_query(string):
    """Valid query either contains title, doi or url."""
    DOI_REGEX = r"10.\d{4,9}\/[-._;()\/:a-zA-Z0-9]+"
    URL_REGEX = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
    REF_REGEX = ref_regex_simple()

    if (mo := re.search(DOI_REGEX, string)) is not None:
        query = mo.group()
        logging.debug(f"{query} is DOI.")
    elif (mo := re.search(URL_REGEX, string)) is not None:
        query = mo.group()
        logging.debug(f"{query} is URL.")
    elif (mo := re.search(REF_REGEX, string)) is not None:
        query = mo.group("title")
        logging.debug(f"{query} is title.")
    else:
        return None

    return query


def valid_fn(path, fn_name):
    """Shorten file name in case it exceeds system's maximum length."""
    PC_PATH_MAX = os.pathconf("/", "PC_PATH_MAX") - 4
    PC_NAME_MAX = os.pathconf("/", "PC_NAME_MAX") - 4
    full_len = len(path + fn_name)

    if full_len > os.pathconf("/", "PC_PATH_MAX"):
        logging.debug("Path too long. Will shorten.")
        fn_name = fn_name[0 : (PC_PATH_MAX - full_len)]
    if len(fn_name) > PC_NAME_MAX:
        logging.debug("File name too long. Will shorten.")
        fn_name = fn_name[0 : (PC_NAME_MAX - len(fn_name))]

    return fn_name


"""Main module."""

import logging
import sys
import click
import requests
from bs4 import BeautifulSoup
import re
from pathlib import Path
import unicodedata
import csv

from .tools import extract_valid_query, valid_fn


class SciHub(object):
    """The SciHub object can be used to download PDFs from SciHub after initialisation."""

    def __init__(self, url, output_path):
        """Initialises the SciHub object with the Sci-Hub url ``url`` and writes all PDFs to the ``output_path`` folder.

        Args:
            url (str): Sci-Hub URL to use
            output_path (Path): The folder to download all PDFs to
        """
        self.url = url
        self.output_path = Path(output_path)
        self.session = requests.Session()

    def download(self, queries):
        """Download articles for provided queries

        Args:
            queries (list(str)): List of queries to look up

        Raises:
            ValueError: If argument is not a string or list of strings
        """
        # make sure queries is of the right format
        if type(queries) == str:
            queries = list(queries)
        elif type(queries) != list:
            raise ValueError("queries argument should be a list or a single string.")

        # get existing downloads or create empty dict for pdf locations
        pdf_paths = self._get_pdf_paths()

        # remove queries that have a valid pdf file already
        queries = self._exclude_existing_queries(queries, pdf_paths)

        try:
            with click.progressbar(queries) as bar:
                for query in bar:
                    try:
                        pdf_path = self._fetch_search(query)
                        pdf_paths[query] = pdf_path
                    except (KeyboardInterrupt, SystemExit) as err:
                        raise err
                    except:
                        logging.error(f"Something went wrong for query: {query}")
                        pdf_paths[query] = ""
        except (KeyboardInterrupt, SystemExit):
            logging.info(
                f"Exiting program. Saving PDF information to {self.output_path}."
            )
        finally:
            self._save_pdf_paths(pdf_paths)

    def _get_pdf_paths(self):
        """Checks for existing pdf_path file or return empty one

        Returns:
            dict: Dictionary containing existing PDFs
        """
        f_path = self.output_path / "pdf_paths.csv"
        pdf_paths = dict()

        if f_path.is_file():
            logging.debug("pdf_paths.csv file detected.")
            with open(f_path, newline="") as f:
                reader = csv.DictReader(f)
                for row in reader:
                    if row["pdf_path"] != "":
                        if Path(row["pdf_path"]).is_file():
                            pdf_paths[row["query"]] = row["pdf_path"]

        return pdf_paths

    def _save_pdf_paths(self, pdf_paths):
        """Saves the queries and corresponding paths to the PDFs after downloading

        Args:
            pdf_paths (dict): Dictionary of paths to the downloaded PDFs
        """
        f_path = self.output_path / "pdf_paths.csv"

        if len(pdf_paths.keys()) > 0:
            with open(f_path, "w") as f:
                w = csv.writer(f)
                w.writerow(["query", "pdf_path"])
                for k, v in pdf_paths.items():
                    w.writerow([k, v])

    def _exclude_existing_queries(self, queries, pdf_paths):
        """Remove queries of which we already have a PDF file

        Args:
            queries (list(str)): List of queries to look up
            pdf_paths (dict): Dictionary of paths to the downloaded PDFs

        Returns:
            list(str): Filtered list of queries
        """
        return [query for query in queries if query not in pdf_paths.keys()]

    def _fetch_search(self, query):
        """Try to find page and return PDF location if succeeded

        Args:
            query (list(str)): List of queries to look up

        Returns:
            str: File location of downloaded PDF corresponding to query
        """
        clean_query = extract_valid_query(query)
        if not clean_query:
            logging.error(
                f"Could not extract valid query from: {query}. Try providing a valid URL, doi or title."
            )
            return None
        else:
            response = self.session.post(self.url, data={"request": clean_query})
            return self._handle_response(response)

    def _handle_response(self, response):
        """Handle a valid response

        Args:
            response (Response): requests.Response object

        Returns:
            str: File location of downloaded PDF corresponding to query
        """
        if response.status_code != 200:
            logging.error(f"Could not connect to Sci-Hub via: {response.url}")
            return None
        else:
            # if status code is okay then transform into beautiful soup
            soup = BeautifulSoup(response.text, features="lxml")
            if self._page_is_valid(soup):
                data = self._extract_data(soup)
                if self._data_is_valid(data):
                    return self._save_pdf(data)

            return None

    def _page_is_valid(self, soup: BeautifulSoup):
        """Sometimes we cannot find the article or we need to solve a CAPTCHA

        Args:
            soup (BeautifulSoup): Soup of the requested search query

        Returns:
            bool: True if shown page is not a missing article or CAPTCHA page
        """
        if re.search(r"article not found", soup.get_text()):
            logging.warn(f"Could not find article.")
            return False
        elif re.search(r"Для просмотра статьи разгадайте капчу", soup.get_text()):
            logging.warn(f"Could not open page due to CAPTCHA.")
            return False
        else:
            return True

    def _extract_data(self, soup: BeautifulSoup):
        """Extract citation, URL and PDF link from page

        Args:
            soup (BeautifulSoup): Soup of the requested search query

        Returns:
            dict: A dictionary containing the citation, URL and PDF link
        """
        URL_REGEX = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
        pdf_url = re.findall(
            URL_REGEX, soup.find("div", id="buttons").select("ul li a")[0]["onclick"]
        )[0][0]

        # if URL does not contain https, then add it
        if not re.match(r"^https://", pdf_url):
            pdf_url = f"https://{pdf_url}"

        return {
            "citation": soup.find("div", id="citation").get_text(),
            "link": soup.find("div", id="link").find("a")["href"],
            "pdf": pdf_url,
        }

    def _data_is_valid(self, data):
        """Check if extracted data contains a valid PDF link

        Args:
            data (dict): A dictionary containing the citation, URL and PDF link

        Returns:
            bool: True if data contains a PDF link
        """
        if data["pdf"] is None:
            return False
        else:
            return True

    def _save_pdf(self, data):
        """Try to download the PDF from the link and save it to the output folder

        Args:
            data (dict): A dictionary containing the citation, URL and PDF link

        Returns:
            str: File location of downloaded PDF corresponding to query
        """
        # open PDF
        response = requests.get(data["pdf"])

        if response.status_code == 200:
            fn_name = unicodedata.normalize("NFKD", data["citation"])
            fn_name = re.sub(r"[^\w\s-]", "", fn_name).strip().lower()
            fn_name = re.sub(r"[-\s]+", "-", fn_name)
            fn_name = valid_fn(str(self.output_path.resolve()), fn_name)
            fn_name = f"{fn_name}.pdf"

            try:
                with open(self.output_path / fn_name, "wb") as pdf:
                    pdf.write(response.content)

                return str(self.output_path.resolve() / fn_name)
            except OSError as err:
                logging.error(err.strerror)
        else:
            logging.error(f"Could not download PDF from: {data['pdf']}")

        return None



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "Click>=7.0",
    "beautifulsoup4>=4.0",
    "requests>=2.0",
]

setup_requirements = []

test_requirements = ["pytest>=6.0"]

setup(
    author="Mark van der Broek",
    author_email="markvanderbroek@gmail.com",
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    description="Command-line and Python API to download PDFs directly from Sci-Hub",
    entry_points={
        "console_scripts": [
            "pyscihub=pyscihub.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="pyscihub",
    name="pyscihub",
    packages=find_packages(include=["pyscihub", "pyscihub.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/markkvdb/pyscihub",
    version="0.1.2",
    zip_safe=False,
)
